

#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import logging
import threading
import queue

from datetime import datetime


from src.base.config import AuraConfig
from src.base.utils import SimpleUtil as SU
from src.core.engine import Engine

from src.scheduling.models import Timeslot, Playlist, PlaylistEntry, PlaylistEntryMetaData
from src.scheduling.api import ApiFetcher



class ProgrammeService():
    """
    The current programme of the calendar. The programme is a set of timeslots for the current day.
    """
    config = None
    logger = None
    programme = None
    last_successful_fetch = None


    def __init__(self):
        """
        Constructor
        """
        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")



    def refresh(self):
        """
        Fetch the latest programme from `AuraCalendarService` which stores it to the database.
        After that, the programme is in turn loaded from the database and stored in `self.programme`.
        """

        # Fetch programme from API endpoints
        self.logger.debug("Trying to fetch new programe from API endpoints...")
        pp = ProgrammePersistence()
        pp_queue = pp.get_queue()
        pp.start() # start fetching thread
        response = pp_queue.get() # wait for the end
        self.logger.debug("... Programme fetch via API done!")

        # Reset last successful fetch state
        lsf = self.last_successful_fetch
        self.last_successful_fetch = None

        if response is None:
            msg = SU.red("Trying to load programme from Engine Database, because AuraCalendarService returned an empty response.")
            self.logger.warning(msg)
        elif type(response) is list:
            self.programme = response
            if self.programme is not None and len(self.programme) > 0:
                self.last_successful_fetch = datetime.now()
                self.logger.info(SU.green("Finished fetching current programme from API"))
            if len(self.programme) == 0:
                self.logger.critical("Programme fetched from Steering/Tank has no entries!")
        elif response.startswith("fetching_aborted"):
            msg = SU.red("Trying to load programme from database only, because fetching was being aborted from AuraCalendarService! Reason: ")
            self.logger.warning(msg + response[16:])
        else:
            msg = SU.red("Trying to load programme from database only, because of an unknown response from AuraCalendarService: " + response)
            self.logger.warning(msg)

        # Always load latest programme from the database
        self.last_successful_fetch = lsf
        self.load_programme_from_db()
        self.logger.info(SU.green("Finished loading current programme from database (%s timeslots)" % str(len(self.programme))))
        for timeslot in self.programme:
            self.logger.debug("\tTimeslot %s with Playlist %s" % (str(timeslot), str(timeslot.playlist)))




    def load_programme_from_db(self):
        """
        Loads the programme from Engine's database and enables them via `self.enable_entries(..)`. After that, the
        current message queue is printed to the console.
        """
        self.programme = Timeslot.get_timeslots()

        if not self.programme:
            self.logger.critical(SU.red("Could not load programme from database. We are in big trouble my friend!"))
            return



    def get_current_entry(self):
        """
        Retrieves the current `PlaylistEntry` which should be played as per programme. 

        Returns:
            (PlaylistEntry): The track which is (or should) currently being played
        """
        now_unix = Engine.engine_time()

        # Load programme if necessary
        if not self.programme:
            self.load_programme_from_db()

        # Check for current timeslot
        current_timeslot = self.get_current_timeslot()
        if not current_timeslot:
            self.logger.warning(SU.red("There's no active timeslot"))
            return None

        # Check for scheduled playlist
        current_playlist = current_timeslot.playlist
        if not current_playlist:
            msg = "There's no playlist assigned to the current timeslot. Most likely a fallback will make things okay again."
            self.logger.warning(SU.red(msg))
            return None

        # Iterate over playlist entries and store the current one
        current_entry = None
        for entry in current_playlist.entries:
            if entry.start_unix <= now_unix and now_unix <= entry.end_unix:
                current_entry = entry
                break
      
        if not current_entry:
            # Nothing playing ... fallback will kick-in
            msg = "There's no entry scheduled for playlist '%s' at %s" % (str(current_playlist), SU.fmt_time(now_unix))
            self.logger.warning(SU.red(msg))
            return None

        return current_entry
 


    def get_current_timeslot(self):
        """
        Retrieves the timeslot currently to be played. 
        
        Returns:
            (Timeslot): The current timeslot
        """
        current_timeslot = None
        now_unix = Engine.engine_time()

        # Iterate over all timeslots and find the one to be played right now
        if self.programme:
            for timeslot in self.programme:
                if timeslot.start_unix <= now_unix and now_unix < timeslot.end_unix:
                    current_timeslot = timeslot
                    break
        
        return current_timeslot



    def get_next_timeslots(self, max_count=0):
        """
        Retrieves the timeslots to be played after the current one.

        Args:
            max_count (Integer): Maximum of timeslots to return, if `0` all exitsing ones are returned

        Returns:
            ([Timeslot]): The next timeslots
        """
        now_unix = Engine.engine_time()
        next_timeslots = []

        for timeslot in self.programme:
            if timeslot.start_unix > now_unix:
                if (len(next_timeslots) < max_count) or max_count == 0:
                    next_timeslots.append(timeslot)
                else:
                    break
                
        return self.filter_scheduling_window(next_timeslots)



    def filter_scheduling_window(self, timeslots):
        """
        Ignore timeslots which are beyond the scheduling window. The end of the scheduling window
        is defined by the config option `scheduling_window_end`. This value defines the seconds
        minus the actual start time of the timeslot.
        """
        now_unix = Engine.engine_time()
        len_before = len(timeslots)
        window_start = self.config.get("scheduling_window_start")
        window_end = self.config.get("scheduling_window_end")
        timeslots = list(filter(lambda s: (s.start_unix - window_end) > now_unix and (s.start_unix - window_start) < now_unix, timeslots))
        len_after = len(timeslots)
        self.logger.info("For now, skipped %s future timeslot(s) which are out of the scheduling window (-%ss <-> -%ss)" % ((len_before - len_after), window_start, window_end))

        return timeslots



    def is_timeslot_in_window(self, timeslot):
        """
        Checks if the timeslot is within the scheduling window.
        """
        now_unix = Engine.engine_time()
        window_start = self.config.get("scheduling_window_start")
        window_end = self.config.get("scheduling_window_end")

        if timeslot.start_unix - window_start < now_unix and \
            timeslot.start_unix - window_end > now_unix:

            return True
        return False




class ProgrammePersistence(threading.Thread):
    """ 
    The `ProgrammePersistence` service retrieves all current schedules and related
    playlists including audio files from the configured API endpoints and stores 
    it in the local database.

    To perform the API queries it utilizes the ApiFetcher class.
    """
    queue = None
    config = None
    logger = None
    fetched_timeslot_data = None
    api_fetcher = None
    stop_event = None


    def __init__(self):
        """
        Initializes the class.
        """
        threading.Thread.__init__(self)

        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")
        self.queue = queue.Queue()
        self.stop_event = threading.Event()
        self.api_fetcher = ApiFetcher()



    def get_queue(self):
        """ 
        Retrieves the queue of fetched data.
        """
        return self.queue



    def run(self):
        """
        Fetch calendar data and store it in the database. Also handles local deletion of remotely
        deleted schedules.

        Returns
            Timeslot ([]):  An arrar of retrieved timeslots passed via `self.queue`
        """
        result = []

        try:
            fetched_timeslot_data = self.api_fetcher.fetch()
            self.logger.debug("Timeslot data fetched from API: " + str(fetched_timeslot_data))

            # If nothing is fetched, return
            if not fetched_timeslot_data:
                self.queue.put("fetching_aborted Nothing fetched")
                return
            
            # Check if existing timeslots have been deleted
            self.update_deleted_timeslots(fetched_timeslot_data)

            # Process fetched timeslots    
            for timeslot in fetched_timeslot_data:

                # Check timeslot for validity
                if "start" not in timeslot:
                    self.logger.warning("No 'start' of timeslot given. Skipping the timeslot: %s " % str(timeslot))
                    continue
                if "end" not in timeslot:
                    self.logger.warning("No 'end' of timeslot given. Skipping the timeslot: %s " % str(timeslot))
                    continue

                # Store the timeslot
                timeslot_db = self.store_timeslot(timeslot)

                # Store playlists to play
                self.store_playlist(timeslot_db, timeslot_db.playlist_id, timeslot["playlist"])
                if timeslot_db.schedule_fallback_id:
                    self.store_playlist(timeslot_db, timeslot_db.schedule_fallback_id, timeslot["schedule_fallback"])
                if timeslot_db.show_fallback_id:
                    self.store_playlist(timeslot_db, timeslot_db.show_fallback_id, timeslot["show_fallback"])
                if timeslot_db.station_fallback_id:
                    self.store_playlist(timeslot_db, timeslot_db.station_fallback_id, timeslot["station_fallback"])

                result.append(timeslot_db)

            # Release the mutex
            self.queue.put(result)
        except Exception as e:
            # Release the mutex
            self.logger.warning("Fetching aborted due to: %s" % str(e), e)
            self.queue.put("fetching_aborted " + str(e))

        # terminate the thread
        return



    def update_deleted_timeslots(self, fetched_timeslot_data):
        """
        Checks if some timeslot has been deleted remotely, so delete it locally too.
        
        Attention: This method has no effect if only a single timeslot got deleted, 
        because this could simply indicate a issue with the API/Steering, since that 
        means no data got retrieved.

        Args:
            fetched_timeslot_data ([dict]): List of timeslot dictionaries from the API
        """
        now_unix = SU.timestamp()
        scheduling_window_start = self.config.get("scheduling_window_start")

        local_timeslots = Timeslot.get_timeslots(datetime.now())
        for local_timeslot in local_timeslots:

            # Ignore timeslots which have already started
            if local_timeslot.start_unix > now_unix:

                # Filter the local timeslot from the fetched ones
                existing_remotely = list(filter(lambda new_timeslot: \
                    new_timeslot["timeslot_id"] == local_timeslot.timeslot_id, fetched_timeslot_data))
                
                if not existing_remotely:
                    # Only allow deletion of timeslots which are deleted before the start of the scheduling window
                    if (local_timeslot.start_unix - scheduling_window_start) > now_unix:                            
                        self.logger.info("Timeslot #%s has been deleted remotely, hence also delete it locally too [%s]" % \
                            (local_timeslot.timeslot_id, str(local_timeslot)))
                        local_timeslot.delete(commit=True)
                        self.logger.info("Remotely deleted timeslot #%s from local database" % local_timeslot.timeslot_id)
                    else:
                        msg = "Timeslot #%s has been deleted remotely. Since the scheduling window has already started, it won't be deleted locally." % local_timeslot.timeslot_id
                        self.logger.error(SU.red(msg))



    def store_timeslot(self, timeslot):
        """
        Stores the given timeslot to the database.

        Args:
            timeslot (Timeslot):    The timeslot
        """
        timeslot_db = Timeslot.for_datetime(timeslot["start"])
        havetoadd = False

        if not timeslot_db:
            self.logger.debug("no timeslot with given timeslot id in database => create new")
            timeslot_db = Timeslot()
            havetoadd = True

        timeslot_db.show_id = timeslot["show_id"]
        timeslot_db.timeslot_id = timeslot["timeslot_id"]
        timeslot_db.timeslot_start = timeslot["start"]
        timeslot_db.timeslot_end = timeslot["end"]
        timeslot_db.show_name = timeslot["show_name"]
        timeslot_db.show_hosts = timeslot["show_hosts"]
        timeslot_db.is_repetition = timeslot["is_repetition"]
        timeslot_db.funding_category = timeslot["show_fundingcategory"]
        timeslot_db.languages = timeslot["show_languages"]
        timeslot_db.type = timeslot["show_type"]
        timeslot_db.category = timeslot["show_categories"]
        timeslot_db.topic = timeslot["show_topics"]
        timeslot_db.musicfocus = timeslot["show_musicfocus"]
        timeslot_db.playlist_id = timeslot["playlist_id"]
        timeslot_db.schedule_fallback_id = timeslot["schedule_fallback_id"]
        timeslot_db.show_fallback_id = timeslot["show_fallback_id"]
        timeslot_db.station_fallback_id = timeslot["station_fallback_id"]

        timeslot_db.store(add=havetoadd, commit=True)
        return timeslot_db



    def store_playlist(self, timeslot_db, playlist_id, fetched_playlist):
        """
        Stores the Playlist to the database.
        """
        if not playlist_id or not fetched_playlist:
            self.logger.debug(f"Playlist ID#{playlist_id} is not available!")
            return
            
        playlist_db = Playlist.select_playlist_for_timeslot(timeslot_db.timeslot_start, playlist_id)
        havetoadd = False

        if not playlist_db:
            playlist_db = Playlist()
            havetoadd = True

        self.logger.debug("Storing playlist %d for timeslot (%s)" % (playlist_id, str(timeslot_db)))
        playlist_db.playlist_id = playlist_id
        playlist_db.timeslot_start = timeslot_db.timeslot_start
        playlist_db.show_name = timeslot_db.show_name
        if "entries" in fetched_playlist:
            playlist_db.entry_count = len(fetched_playlist["entries"])
        else:
            playlist_db.entry_count = 0

        playlist_db.store(havetoadd, commit=True)
      
        if playlist_db.entry_count > 0:
            self.store_playlist_entries(timeslot_db, playlist_db, fetched_playlist)

        return playlist_db



    def store_playlist_entries(self, timeslot_db, playlist_db, fetched_playlist):
        """
        Stores the playlist entries to the database.
        """
        entry_num = 0
        time_marker = playlist_db.start_unix

        self.expand_entry_duration(timeslot_db, fetched_playlist)
        self.delete_orphaned_entries(playlist_db, fetched_playlist)  

        for entry in fetched_playlist["entries"]:
            entry_db = PlaylistEntry.select_playlistentry_for_playlist(playlist_db.artificial_id, entry_num)
            havetoadd = False
            if not entry_db:
                entry_db = PlaylistEntry()
                havetoadd = True

            entry_db.entry_start = datetime.fromtimestamp(time_marker)
            entry_db.artificial_playlist_id = playlist_db.artificial_id
            entry_db.entry_num = entry_num
            entry_db.duration = SU.nano_to_seconds(entry["duration"])

            if "uri" in entry:
                # FIXME Refactor mix of uri/filename/file/source
                entry_db.uri = entry["uri"]
                entry_db.source = entry["uri"]
            if "filename" in entry:
                entry_db.source = entry["filename"]

            entry_db.store(havetoadd, commit=True)

            if "file" in entry:
                self.store_playlist_entry_metadata(entry_db, entry["file"]["metadata"])

            entry_num = entry_num + 1
            time_marker += entry_db.duration



    def delete_orphaned_entries(self, playlist_db, fetched_playlist):
        """
        Deletes all playlist entries which are beyond the current playlist's `entry_count`.
        Such entries might be existing due to a remotely changed playlist, which now has
        less entries than before.
        """
        new_last_idx = len(fetched_playlist["entries"])
        existing_last_idx = PlaylistEntry.count_entries(playlist_db.artificial_id)-1

        if existing_last_idx < new_last_idx:
            return 

        for entry_num in range(new_last_idx, existing_last_idx+1, 1):
            PlaylistEntry.delete_entry(playlist_db.artificial_id, entry_num)            
            self.logger.info(SU.yellow("Deleted playlist entry %s:%s" % (playlist_db.artificial_id, entry_num)))
            entry_num += 1


    def expand_entry_duration(self, timeslot_db, fetched_playlist):
        """
        If some playlist entry doesn't have a duration assigned, its duration is expanded to the
        remaining duration of the playlist (= timeslot duration minus playlist entries with duration).
        If there's more than one entry without duration, such entries are removed from the playlist.
        """
        total_seconds = (timeslot_db.timeslot_end - timeslot_db.timeslot_start).total_seconds()
        total_duration = SU.seconds_to_nano(total_seconds)
        actual_duration = 0        
        missing_duration = []
        idx = 0

        for entry in fetched_playlist["entries"]:
            if not "duration" in entry:
                missing_duration.append(idx)
            else:
                actual_duration += entry["duration"]
            idx += 1
                
        if len(missing_duration) == 1:
            fetched_playlist["entries"][missing_duration[0]]["duration"] = total_duration - actual_duration
            self.logger.info("Expanded duration of playlist entry #%s:%s" % (fetched_playlist["id"], missing_duration[0]))

        elif len(missing_duration) > 1:
            # This case should actually never happen, as TANK doesn't allow more than one entry w/o duration anymore
            for i in reversed(missing_duration[1:-1]):
                self.logger.error(SU.red("Deleted Playlist Entry without duration: %s" % \
                    str(fetched_playlist["entries"][i])))
                del fetched_playlist["entries"][i]



    def store_playlist_entry_metadata(self, entry_db, metadata):
        """
        Stores the meta-data for a PlaylistEntry.
        """
        metadata_db = PlaylistEntryMetaData.select_metadata_for_entry(entry_db.artificial_id)
        havetoadd = False
        if not metadata_db:
            metadata_db = PlaylistEntryMetaData()
            havetoadd = True

        metadata_db.artificial_entry_id = entry_db.artificial_id

        if "artist" in metadata:
            metadata_db.artist = metadata["artist"]
        else:
            metadata_db.artist = ""
        
        if "album" in metadata:
            metadata_db.album = metadata["album"]
        else:
            metadata_db.album = ""

        if "title" in metadata:
            metadata_db.title = metadata["title"]
        else:
            metadata_db.title = ""

        metadata_db.store(havetoadd, commit=True)



    def stop(self):
        self.stop_event.set()
